from operator import itemgetter

from ..utils import update_available
from .baseresource import BaseResource
from .bettingresources import (
    MarketBook,
    CurrentOrders,
)
from ..enums import (
    StreamingOrderType,
    StreamingPersistenceType,
    StreamingSide,
    StreamingStatus,
)


class MarketDefinitionRunner(BaseResource):
    """
    :type adjustment_factor: float
    :type id: int
    :type removal_date: datetime.datetime
    :type sort_priority: int
    :type status: unicode
    """
    adjustment_factor = None
    id = None
    removal_date = None
    sort_priority = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'runners'
        attributes = {
            'id': 'id',
            'sortPriority': 'sort_priority',
            'status': 'status',
            'adjustmentFactor': 'adjustment_factor',
            'removalDate': 'removal_date',
        }
        datetime_attributes = (
            'removalDate',
        )


class MarketDefinition(BaseResource):
    """
    :type bet_delay: int
    :type betting_type: unicode
    :type bsp_market: bool
    :type bsp_reconciled: bool
    :type complete: bool
    :type country_code: unicode
    :type cross_matching: bool
    :type discount_allowed: bool
    :type event_id: unicode
    :type event_type_id: unicode
    :type in_play: bool
    :type market_base_rate: int
    :type market_time: datetime.datetime
    :type market_type: unicode
    :type number_of_active_runners: int
    :type number_of_winners: int
    :type open_date: datetime.datetime
    :type persistence_enabled: bool
    :type regulators: list[unicode]
    :type runners: list[MarketDefinitionRunner]
    :type runners_voidable: bool
    :type settled_time: datetime.datetime
    :type status: unicode
    :type suspend_time: datetime.datetime
    :type timezone: unicode
    :type turn_in_play_enabled: bool
    :type venue: unicode
    :type version: int
    """
    bet_delay = None
    betting_type = None
    bsp_market = None
    bsp_reconciled = None
    complete = None
    country_code = None
    cross_matching = None
    discount_allowed = None
    event_id = None
    event_type_id = None
    in_play = None
    market_base_rate = None
    market_time = None
    market_type = None
    number_of_active_runners = None
    number_of_winners = None
    open_date = None
    persistence_enabled = None
    regulators = None
    runners = None
    runners_voidable = None
    settled_time = None
    status = None
    suspend_time = None
    timezone = None
    turn_in_play_enabled = None
    venue = None
    version = None

    class Meta(BaseResource.Meta):
        identifier = 'market_definition'
        attributes = {
            'betDelay': 'bet_delay',
            'bettingType': 'betting_type',
            'bspMarket': 'bsp_market',
            'bspReconciled': 'bsp_reconciled',
            'complete': 'complete',
            'countryCode': 'country_code',
            'crossMatching': 'cross_matching',
            'discountAllowed': 'discount_allowed',
            'eventId': 'event_id',
            'eventTypeId': 'event_type_id',
            'inPlay': 'in_play',
            'marketBaseRate': 'market_base_rate',
            'marketTime': 'market_time',
            'marketType': 'market_type',
            'numberOfActiveRunners': 'number_of_active_runners',
            'numberOfWinners': 'number_of_winners',
            'persistenceEnabled': 'persistence_enabled',
            'regulators': 'regulators',
            'runnersVoidable': 'runners_voidable',
            'settledTime': 'settled_time',
            'openDate': 'open_date',
            'status': 'status',
            'suspendTime': 'suspend_time',
            'venue': 'venue',
            'timezone': 'timezone',
            'turnInPlayEnabled': 'turn_in_play_enabled',
            'version': 'version',
        }
        sub_resources = {
            'runners': MarketDefinitionRunner
        }
        datetime_attributes = (
            'marketTime',
            'openDate',
            'suspendTime',
            'settledTime',
        )


class RunnerBook(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'runners'
        attributes = {
            'id': 'selection_id',
            'ltp': 'last_price_traded',
            'tv': 'total_matched',
            'trd': 'traded',
            'atb': 'available_to_back',
            'batb': 'best_available_to_back',
            'bdatb': 'best_display_available_to_back',
            'atl': 'available_to_lay',
            'batl': 'best_available_to_lay',
            'bdatl': 'best_display_available_to_lay',
            'spn': 'starting_price_near',
            'spf': 'starting_price_far',
            'spb': 'starting_price_back',
            'spl': 'starting_price_lay',
        }

    def update_traded(self, traded_update):
        """:param traded_update: price, size
        """
        if not traded_update:
            self.traded = traded_update
        elif not self.traded:
            self.traded = traded_update
        else:
            update_available(self.traded, traded_update, 1)

    def update_available_to_back(self, book_update):
        """:param book_update: price, size
        """
        if not self.available_to_back:
            self.available_to_back = book_update
        else:
            update_available(self.available_to_back, book_update, 1)

    def update_available_to_lay(self, book_update):
        """:param book_update: price, size
        """
        if not self.available_to_lay:
            self.available_to_lay = book_update
        else:
            update_available(self.available_to_lay, book_update, 1)

    def update_best_available_to_back(self, book_update):
        """:param book_update: level, price, size
        """
        if not self.best_available_to_back:
            self.best_available_to_back = book_update
        else:
            update_available(self.best_available_to_back, book_update, 2)

    def update_best_available_to_lay(self, book_update):
        """:param book_update: level, price, size
        """
        if not self.best_available_to_lay:
            self.best_available_to_lay = book_update
        else:
            update_available(self.best_available_to_lay, book_update, 2)

    def update_best_display_available_to_back(self, book_update):
        """:param book_update: level, price, size
        """
        if not self.best_display_available_to_back:
            self.best_display_available_to_back = book_update
        else:
            update_available(self.best_display_available_to_back, book_update, 2)

    def update_best_display_available_to_lay(self, book_update):
        """:param book_update: level, price, size
        """
        if not self.best_display_available_to_lay:
            self.best_display_available_to_lay = book_update
        else:
            update_available(self.best_display_available_to_lay, book_update, 2)

    def update_starting_price_back(self, book_update):
        """:param book_update: price, size
        """
        if not self.starting_price_back:
            self.starting_price_back = book_update
        else:
            update_available(self.starting_price_back, book_update, 1)

    def update_starting_price_lay(self, book_update):
        """:param book_update: price, size
        """
        if not self.starting_price_lay:
            self.starting_price_lay = book_update
        else:
            update_available(self.starting_price_lay, book_update, 1)

    @property
    def serialise_traded_volume(self):
        if self.traded:
            return [{'price': volume[0], 'size': volume[1]}
                    for volume in sorted(self.traded, key=itemgetter(0))]
        else:
            return []

    @property
    def serialise_available_to_back(self):
        if self.available_to_back:
            return [{'price': volume[0], 'size': volume[1]}
                    for volume in sorted(self.available_to_back, key=itemgetter(0), reverse=True)]
        elif self.best_display_available_to_back:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_display_available_to_back, key=itemgetter(0))]
        elif self.best_available_to_back:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_available_to_back, key=itemgetter(0))]
        else:
            return []

    @property
    def serialise_available_to_lay(self):
        if self.available_to_lay:
            return [{'price': volume[0], 'size': volume[1]}
                    for volume in sorted(self.available_to_lay, key=itemgetter(0))]
        elif self.best_display_available_to_lay:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_display_available_to_lay, key=itemgetter(0))]
        elif self.best_available_to_lay:
            return [{'price': volume[1], 'size': volume[2]}
                    for volume in sorted(self.best_available_to_lay, key=itemgetter(0))]
        return []

    def serialise(self, status):
        return {
            'status': status,
            'ex': {
                'tradedVolume': self.serialise_traded_volume,
                'availableToBack': self.serialise_available_to_back,
                'availableToLay': self.serialise_available_to_lay
            },
            'adjustmentFactor': None,
            'lastPriceTraded': self.last_price_traded,
            'handicap': None,
            'totalMatched': self.total_matched,
            'selectionId': self.selection_id
        }


class MarketBookCache(BaseResource):

    class Meta(BaseResource.Meta):
        identifier = 'market_book_cache'
        attributes = {
            'id': 'market_id',
            'img': 'image',
            'tv': 'total_matched'
        }
        sub_resources = {
            'marketDefinition': MarketDefinition,
            'rc': RunnerBook
        }

    def update_cache(self, market_change, publish_time):
        self._datetime_updated = self.strip_datetime(publish_time)
        self.publish_time = publish_time

        market_definition = market_change.get('marketDefinition')
        if market_definition:
            self.market_definition = MarketDefinition(**market_definition)

        traded_volume = market_change.get('tv')
        if traded_volume:
            self.total_matched = traded_volume

        runner_change = market_change.get('rc')
        if runner_change:
            for new_data in runner_change:
                selection_id = new_data.get('id')
                runner = self.runner_dict.get(selection_id)
                if runner:
                    if new_data.get('ltp'):
                        runner.last_price_traded = new_data.get('ltp')
                    if new_data.get('tv'):
                        runner.total_matched = new_data.get('tv')
                    if new_data.get('spn'):
                        runner.spn = new_data.get('spn')
                    if new_data.get('spf'):
                        runner.spf = new_data.get('spf')
                    if new_data.get('trd'):
                        runner.update_traded(new_data.get('trd'))
                    if new_data.get('atb'):
                        runner.update_available_to_back(new_data.get('atb'))
                    if new_data.get('atl'):
                        runner.update_available_to_lay(new_data.get('atl'))
                    if new_data.get('batb'):
                        runner.update_best_available_to_back(new_data.get('batb'))
                    if new_data.get('batl'):
                        runner.update_best_available_to_lay(new_data.get('batl'))
                    if new_data.get('bdatb'):
                        runner.update_best_display_available_to_back(new_data.get('bdatb'))
                    if new_data.get('bdatl'):
                        runner.update_best_display_available_to_lay(new_data.get('bdatl'))
                    if new_data.get('spb'):
                        runner.update_starting_price_back(new_data.get('spb'))
                    if new_data.get('spl'):
                        runner.update_starting_price_back(new_data.get('spl'))
                else:
                    self.runners.append(RunnerBook(**new_data))

    def create_market_book(self, unique_id):
        return MarketBook(
            date_time_sent=self._datetime_updated,
            streaming_unique_id=unique_id,
            market_definition=self.market_definition,
            **self.serialise
        )

    @property
    def runner_dict(self):
        return {runner.selection_id: runner for runner in self.runners}

    @property
    def market_definition_dict(self):
        return {runner.id: runner for runner in self.market_definition.runners}

    @property
    def serialise(self):
        """Creates standard market book json response,
        will error if EX_MARKET_DEF not incl.
        """
        return {
            'marketId': self.market_id,
            'totalAvailable': None,
            'betDelay': self.market_definition.bet_delay,
            'version': self.market_definition.version,
            'complete': self.market_definition.complete,
            'numberOfRunners': None,
            'runnersVoidable': self.market_definition.runners_voidable,
            'totalMatched': self.total_matched,
            'status': self.market_definition.status,
            'bspReconciled': self.market_definition.bsp_reconciled,
            'isMarketDataDelayed': None,
            'lastMatchTime': None,
            'crossMatching': self.market_definition.cross_matching,
            'inplay': self.market_definition.in_play,
            'numberOfWinners': self.market_definition.number_of_winners,
            'numberOfActiveRunners': self.market_definition.number_of_active_runners,
            'runners': [runner.serialise(self.market_definition_dict.get(runner.selection_id).status)
                        for runner in self.runners],
            'publishTime': self.publish_time,
        }


class UnmatchedOrder(BaseResource):

    class Meta(BaseResource.Meta):
        identifier = 'unmatched_orders'
        attributes = {
            'id': 'bet_id',
            'p': 'price',
            's': 'size',
            'bsp': 'bsp_liability',
            'side': 'side',
            'status': 'status',
            'pt': 'persistence_type',
            'ot': 'order_type',
            'pd': 'placed_date',
            'md': 'matched_date',
            'avp': 'average_price_matched',
            'sm': 'size_matched',
            'sr': 'size_remaining',
            'sl': 'size_lapsed',
            'sc': 'size_cancelled',
            'sv': 'size_voided',
            'rac': 'regulator_auth_code',
            'rc': 'regulator_code',
            'rfo': 'reference_order',
            'rfs': 'reference_strategy',
        }
        datetime_attributes = (
            'pd',
            'md'
        )

    def serialise(self, market_id, selection_id):
        return {
            'averagePriceMatched': self.average_price_matched or 0.0,
            'betId': self.bet_id,
            'bspLiability': self.bsp_liability,
            'handicap': 0.0,
            'marketId': market_id,
            'matchedDate': self.matched_date.strftime('%Y-%m-%dT%H:%M:%S.%fZ') if self.matched_date is not None else self.matched_date,
            'orderType': StreamingOrderType[self.order_type].value,
            'persistenceType': StreamingPersistenceType[self.persistence_type].value,
            'placedDate': self.placed_date.strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
            'priceSize': {
                'price': self.price,
                'size': self.size
            },
            'regulatorCode': self.regulator_code,
            'selectionId': selection_id,
            'side': StreamingSide[self.side].value,
            'sizeCancelled': self.size_cancelled,
            'sizeLapsed': self.size_lapsed,
            'sizeMatched': self.size_matched,
            'sizeRemaining': self.size_remaining,
            'sizeVoided': self.size_voided,
            'status': StreamingStatus[self.status].value,
            'customerStrategyRef': self.reference_strategy,
            'customerOrderRef': self.reference_order,
        }


class OrderBookRunner(BaseResource):

    class Meta(BaseResource.Meta):
        identifier = 'runners'
        attributes = {
            'id': 'selection_id',
            'ml': 'matched_lays',
            'mb': 'matched_backs',
        }
        sub_resources = {
            'uo': UnmatchedOrder
        }

    def update_matched_backs(self, matched_backs):
        if not self.matched_backs:
            self.matched_backs = [matched_back for matched_back in matched_backs]
        else:
            update_available(self.matched_backs, matched_backs, 1)

    def update_matched_lays(self, matched_lays):
        if not self.matched_lays:
            self.matched_lays = [matched_lay for matched_lay in matched_lays]
        else:
            update_available(self.matched_lays, matched_lays, 1)

    def update_unmatched(self, unmatched_orders):
        order_dict = {order.bet_id: order for order in self.unmatched_orders}
        for unmatched_order in unmatched_orders:
            if unmatched_order.get('id') in order_dict:
                for n, order in enumerate(self.unmatched_orders):
                    if order.bet_id == unmatched_order.get('id'):
                        self.unmatched_orders[n] = UnmatchedOrder(**unmatched_order)
                        break
            else:
                self.unmatched_orders.append(UnmatchedOrder(**unmatched_order))

    def serialise_orders(self, market_id):
        return [order.serialise(market_id, self.selection_id) for order in self.unmatched_orders]


class OrderBookCache(BaseResource):

    class Meta(BaseResource.Meta):
        identifier = 'order_book_cache'
        attributes = {
            'id': 'market_id',
            'closed': 'closed'
        }
        sub_resources = {
            'orc': OrderBookRunner
        }

    def update_cache(self, order_book, publish_time):
        self._datetime_updated = self.strip_datetime(publish_time)

        for order_changes in order_book.get('orc', []):
            selection_id = order_changes.get('id')
            runner = self.runner_dict.get(selection_id)
            if runner:
                runner.update_matched_lays(order_changes.get('ml', []))
                runner.update_matched_backs(order_changes.get('mb', []))
                runner.update_unmatched(order_changes.get('uo', []))
            else:
                self.runners.append(OrderBookRunner(**order_changes))

    def create_order_book(self, unique_id):
        return CurrentOrders(date_time_sent=self._datetime_updated, streaming_unique_id=unique_id, **self.serialise)

    @property
    def runner_dict(self):
        return {runner.selection_id: runner for runner in self.runners}

    @property
    def serialise(self):
        orders = []
        for runner in self.runners:
            orders.extend(runner.serialise_orders(self.market_id))
        return {
            'currentOrders': orders,
            'moreAvailable': False
        }



import datetime

from .baseendpoint import BaseEndpoint
from .. import resources


class Account(BaseEndpoint):

    URI = 'AccountAPING/v1.0/'
    connect_timeout = 6.05

    def get_account_funds(self, params=None, session=None):
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'getAccountFunds')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.AccountFunds, date_time_sent)

    def get_account_details(self, params=None, session=None):
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'getAccountDetails')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.AccountDetails, date_time_sent)

    def get_account_statement(self, params=None, session=None):
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'getAccountStatement')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.AccountStatementResult, date_time_sent)

    def list_currency_rates(self, params=None, session=None):
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listCurrencyRates')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.CurrencyRate, date_time_sent)

    def transfer_funds(self, params=None, session=None):
        raise DeprecationWarning('As of 20/09/2016 AUS wallet has been removed, function still available for when '
                                 'accounts are added in 2017.')
        # date_time_sent = datetime.datetime.utcnow()
        # method = '%s%s' % (self.URI, 'transferFunds')
        # response = self.request(method, params, session)
        # return self.process_response(response.json(), resources.TransferFunds, date_time_sent)

    @property
    def url(self):
        return '%s%s' % (self.client.api_uri, 'account/json-rpc/v1')



from .baseresource import BaseResource


class InningOne(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'inning_one'
        attributes = {
            'overs': 'overs',
            'runs': 'runs',
            'wickets': 'wickets'
        }


class Home(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'home'
        attributes = {
            'bookingPoints': 'booking_points',
            'fullTimeScore': 'full_time_score',
            'games': 'games',
            'halfTimeScore': 'half_time_score',
            'name': 'name',
            'numberOfCards': 'number_of_cards',
            'numberOfCorners': 'number_of_corners',
            'numberOfCornersFirstHalf': 'number_of_corners_first_half',
            'numberOfCornersSecondHalf': 'number_of_corners_second_half',
            'numberOfRedCards': 'number_of_red_cards',
            'numberOfYellowCards': 'number_of_yellow_cards',
            'penaltiesScore': 'penalties_score',
            'penaltiesSequence': 'penalties_sequence',
            'score': 'score',
            'sets': 'sets',
            'highlight': 'highlight',
            'aces': 'aces',
            'doubleFaults': 'double_faults',
            'gameSequence': 'game_sequence',
            'isServing': 'is_serving',
            'playerSeed': 'player_seed',
            'serviceBreaks': 'service_breaks',
        }
        sub_resources = {
            'inning1': InningOne
        }


class Away(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'away'
        attributes = {
            'bookingPoints': 'booking_points',
            'fullTimeScore': 'full_time_score',
            'games': 'games',
            'halfTimeScore': 'half_time_score',
            'name': 'name',
            'numberOfCards': 'number_of_cards',
            'numberOfCorners': 'number_of_corners',
            'numberOfCornersFirstHalf': 'number_of_corners_first_half',
            'numberOfCornersSecondHalf': 'number_of_corners_second_half',
            'numberOfRedCards': 'number_of_red_cards',
            'numberOfYellowCards': 'number_of_yellow_cards',
            'penaltiesScore': 'penalties_score',
            'penaltiesSequence': 'penalties_sequence',
            'score': 'score',
            'sets': 'sets',
            'highlight': 'highlight',
            'aces': 'aces',
            'doubleFaults': 'double_faults',
            'gameSequence': 'game_sequence',
            'isServing': 'is_serving',
            'playerSeed': 'player_seed',
            'serviceBreaks': 'service_breaks',
        }
        sub_resources = {
            'inning1': InningOne
        }


class Score(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'score'
        attributes = {
            'bookingPoints': 'booking_points',
            'numberOfCards': 'number_of_cards',
            'numberOfCorners': 'number_of_corners',
            'numberOfCornersFirstHalf': 'number_of_corners_first_half',
            'numberOfCornersSecondHalf': 'number_of_corners_second_half',
            'numberOfRedCards': 'number_of_red_cards',
            'numberOfYellowCards': 'number_of_yellow_cards'

        }
        sub_resources = {
            'away': Away,
            'home': Home,
        }


class UpdateDetail(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'update_detail'
        attributes = {
            'elapsedRegularTime': 'elapsed_regular_time',
            'matchTime': 'match_time',
            'type': 'type',
            'updateId': 'update_id',
            'updateTime': 'update_time',
            'updateType': 'update_type',
            'team': 'team',
            'teamName': 'team_name',
        }
        datetime_attributes = (
            'updateTime'
        )


class EventTimeline(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'event_timeline'
        attributes = {
            'eventId': 'event_id',
            'elapsedRegularTime': 'elapsed_regular_time',
            'eventTypeId': 'event_type_id',
            'inPlayMatchStatus': 'in_play_match_status',
            'status': 'status',
            'timeElapsed': 'time_elapsed'

        }
        sub_resources = {
            'score': Score,
            'updateDetails': UpdateDetail,
        }


class FullTimeElapsed(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'full_time_elapsed'
        attributes = {
            'hour': 'hour',
            'min': 'min',
            'sec': 'sec'
        }


class StateOfBall(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'state_of_ball'
        attributes = {
            'appealId': 'appeal_id',
            'appealTypeName': 'appeal_type_name',
            'batsmanName': 'batsman_name',
            'batsmanRuns': 'batsman_runs',
            'bowlerName': 'bowler_name',
            'bye': 'bye',
            'dismissalTypeName': 'dismissal_type_name',
            'legBye': 'leg_bye',
            'noBall': 'no_ball',
            'outcomeId': 'outcome_id',
            'overBallNumber': 'over_ball_number',
            'overNumber': 'over_number',
            'referralOutcome': 'referral_outcome',
            'wide': 'wide'
        }


class Scores(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'scores'
        attributes = {
            'eventId': 'event_id',
            'elapsedRegularTime': 'elapsed_regular_time',
            'elapsedAddedTime': 'elapsed_added_time',
            'eventTypeId': 'event_type_id',
            'matchStatus': 'match_status',
            'timeElapsed': 'time_elapsed',
            'timeElapsedSeconds': 'time_elapsed_seconds',
            'status': 'status',
            'currentDay': 'current_day',
            'currentSet': 'current_set',
            'description': 'description',
            'matchType': 'match_type',
            'currentGame': 'current_game',
            'currentPoint': 'current_point'
        }
        sub_resources = {
            'fullTimeElapsed': FullTimeElapsed,
            'score': Score,
            'stateOfBall': StateOfBall
        }



import datetime
from requests import ConnectionError

from .baseendpoint import BaseEndpoint
from ..resources import KeepAliveResource
from ..exceptions import KeepAliveError, APIError
from ..utils import check_status_code


class KeepAlive(BaseEndpoint):

    _error = KeepAliveError

    def __call__(self, session=None):
        date_time_sent = datetime.datetime.utcnow()
        response = self.request(session=session)
        response_json = response.json()
        self.client.set_session_token(response_json.get('token'))
        return self.process_response(response_json, KeepAliveResource, date_time_sent)

    def request(self, payload=None, params=None, session=None):
        session = session or self.client.session
        try:
            response = session.post(self.url, headers=self.client.keep_alive_headers, cert=self.client.cert)
        except ConnectionError:
            raise APIError(None, exception='ConnectionError')
        except Exception as e:
            raise APIError(None, exception=e)

        check_status_code(response)
        if self._error_handler:
            self._error_handler(response.json())
        return response

    def _error_handler(self, response, method=None, params=None):
        if response.get('status') != 'SUCCESS':
            raise self._error(response)

    @property
    def url(self):
        return '%s%s' % (self.client.identity_uri, 'keepAlive')



from .baseresource import BaseResource


class LoginResource(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'login'
        attributes = {
            'sessionToken': 'session_token',
            'loginStatus': 'login_status',
        }


class KeepAliveResource(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'keep_alive'
        attributes = {
            'product': 'product',
            'status': 'status',
            'token': 'token',
            'error': 'error',
        }


class LogoutResource(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'logout'
        attributes = {
            'product': 'product',
            'status': 'status',
            'token': 'token',
            'error': 'error',
        }



import re
import datetime
from requests import ConnectionError

from ..exceptions import APIError, RaceCardError
from ..utils import check_status_code
from .baseendpoint import BaseEndpoint
from .. import resources


class RaceCard(BaseEndpoint):

    app_key = None

    def login(self, session=None):
        session = session or self.client.session
        response = session.get(self.login_url)
        app_key = re.findall(r'''"appKey":\s"(.*?)"''', response.text)
        if app_key:
            self.app_key = app_key[0]
        else:
            raise RaceCardError("Unable to find appKey")

    def get_race_card(self, market_ids, data_entries=None, session=None):
        """
        :rtype: list[resources.RaceCard]
        """
        if not self.app_key:
            raise RaceCardError("You need to login before requesting a race_card\n"
                                "APIClient.race_card.login()")
        date_time_sent = datetime.datetime.utcnow()
        response = self.request(method=market_ids, params=data_entries, session=session)
        return self.process_response(response.json(), resources.RaceCard, date_time_sent)

    def request(self, method=None, params=None, session=None):
        session = session or self.client.session
        try:
            response = session.get(self.url, params=self.create_req(method, params),
                                   headers=self.headers)
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)

        check_status_code(response)
        return response

    @staticmethod
    def create_req(method=None, params=None):
        if not params:
            params = "RACE, TIMEFORM_DATA, RUNNERS, RUNNER_DETAILS"
        return {
            'dataEntries': params,
            'marketId': ','.join(method)
        }

    @property
    def headers(self):
        return {
            'Connection': 'keep-alive',
            'Content-Type': 'application/json',
            'X-Application': self.app_key
        }

    @property
    def login_url(self):
        return 'https://www.betfair.com/exchange/plus/'

    @property
    def url(self):
        return 'https://www.betfair.com/rest/v2/raceCard'



from .apiclient import APIClient
from .exceptions import BetfairError
from .streaming import StreamListener
from .filters import MarketFilter, StreamingMarketFilter, StreamingMarketDataFilter


__title__ = 'betfairlightweight'
__version__ = '0.9.10'
__author__ = 'Liam Pauling'





class BaseFilter:
    """Base Filter"""


class MarketFilter(BaseFilter):

    def __init__(self, market_ids=None, bsp_only=None, market_betting_types=None, event_type_ids=None, event_ids=None,
                 turn_in_play_enabled=None, market_type_codes=None, venues=None, market_countries=None, text_query=None,
                 competition_ids=None, in_play_only=None, market_start_time=None, with_orders=None):
        self.text_query = text_query
        self.event_type_ids = event_type_ids or []
        self.event_ids = event_ids or []
        self.competition_ids = competition_ids or []
        self.market_ids = market_ids or []
        self.venues = venues or []
        self.bsp_only = bsp_only
        self.turn_in_play_enabled = turn_in_play_enabled
        self.in_play_only = in_play_only
        self.market_betting_types = market_betting_types or []
        self.market_type_codes = market_type_codes or []
        self.market_countries = market_countries or []
        self.market_start_time = market_start_time
        self.with_orders = with_orders or []

    @property
    def serialise(self):
        return {
            'marketIds': self.market_ids,
            'textQuery': self.text_query,
            'marketBettingTypes': self.market_betting_types,
            'eventTypeIds': self.event_type_ids,
            'eventIds': self.event_ids,
            'turnInPlayEnabled': self.turn_in_play_enabled,
            'inPlayOnly': self.in_play_only,
            'marketTypeCodes': self.market_type_codes,
            'venues': self.venues,
            'marketCountries': self.market_countries,
            'bspOnly': self.bsp_only,
            'competitionIds': self.competition_ids,
            'marketStartTime': self.market_start_time,
            'withOrders': self.with_orders,
        }


class StreamingMarketFilter(BaseFilter):

    def __init__(self, market_ids=None, bsp_market=None, betting_types=None, event_type_ids=None, event_ids=None,
                 turn_in_play_enabled=None, market_types=None, venues=None, country_codes=None):
        self.market_ids = market_ids or []
        self.bsp_market = bsp_market
        self.betting_types = betting_types or []
        self.event_type_ids = event_type_ids or []
        self.event_ids = event_ids or []
        self.turn_in_play_enabled = turn_in_play_enabled
        self.market_types = market_types or []
        self.venues = venues or []
        self.country_codes = country_codes or []

    @property
    def serialise(self):
        return {
            'marketIds': self.market_ids,
            'bspMarket': self.bsp_market,
            'bettingTypes': self.betting_types,
            'eventTypeIds': self.event_type_ids,
            'eventIds': self.event_ids,
            'turnInPlayEnabled': self.turn_in_play_enabled,
            'marketTypes': self.market_types,
            'venues': self.venues,
            'countryCodes': self.country_codes,
        }


class StreamingMarketDataFilter(BaseFilter):

    def __init__(self, fields=None, ladder_levels=None):
        """
        fields: EX_BEST_OFFERS_DISP, EX_BEST_OFFERS, EX_ALL_OFFERS, EX_TRADED,
                EX_TRADED_VOL, EX_LTP, EX_MARKET_DEF, SP_TRADED, SP_PROJECTED
        ladder_levels: 1->10
        """
        self.fields = fields or []
        self.ladder_levels = ladder_levels

    @property
    def serialise(self):
        return {
            'fields': self.fields,
            'ladderLevels': self.ladder_levels
        }



from .baseresource import BaseResource


class EventType(BaseResource):
    """
    :type id: unicode
    :type name: unicode
    """
    id = None
    name = None

    class Meta(BaseResource.Meta):
        identifier = 'event_type'
        attributes = {
            'id': 'id',
            'name': 'name'
        }


class EventTypeResult(BaseResource):
    """
    :type event_type: EventType
    :type market_count: int
    """
    event_type = None
    market_count = None

    class Meta(BaseResource.Meta):
        identifier = 'event_type_result'
        attributes = {
            'marketCount': 'market_count'
        }
        sub_resources = {
            'eventType': EventType
        }


class Competition(BaseResource):
    """
    :type id: unicode
    :type name: unicode
    """
    id = None
    name = None

    class Meta(BaseResource.Meta):
        identifier = 'competition'
        attributes = {
            'id': 'id',
            'name': 'name'
        }


class CompetitionResult(BaseResource):
    """
    :type competition: Competition
    :type competition_region: unicode
    :type market_count: int
    """
    competition = None
    competition_region = None
    market_count = None

    class Meta(BaseResource.Meta):
        identifier = 'competition_result'
        attributes = {
            'marketCount': 'market_count',
            'competitionRegion': 'competition_region'
        }
        sub_resources = {
            'competition': Competition
        }


class TimeRange(BaseResource):
    """
    :type _from: datetime.datetime
    :type to: datetime.datetime
    """
    _from = None
    to = None

    class Meta(BaseResource.Meta):
        identifier = 'time_range'
        attributes = {
            'from': '_from',
            'to': 'to'
        }
        datetime_attributes = (
            'from',
            'to'
        )


class TimeRangeResult(BaseResource):
    """
    :type market_count: int
    :type time_range: TimeRange
    """
    market_count = None
    time_range = None

    class Meta(BaseResource.Meta):
        identifier = 'time_range_result'
        attributes = {
            'marketCount': 'market_count'
        }
        sub_resources = {
            'timeRange': TimeRange
        }


class Event(BaseResource):
    """
    :type country_code: unicode
    :type id: unicode
    :type name: unicode
    :type open_date: datetime.datetime
    :type time_zone: unicode
    :type venue: unicode
    """
    country_code = None
    id = None
    name = None
    open_date = None
    time_zone = None
    venue = None

    class Meta(BaseResource.Meta):
        identifier = 'event'
        attributes = {
            'id': 'id',
            'openDate': 'open_date',
            'timezone': 'time_zone',
            'countryCode': 'country_code',
            'name': 'name',
            'venue': 'venue'
        }
        datetime_attributes = (
            'openDate'
        )


class EventResult(BaseResource):
    """
    :type event: Event
    :type market_count: int
    """
    event = None
    market_count = None

    class Meta(BaseResource.Meta):
        identifier = 'event_result'
        attributes = {
            'marketCount': 'market_count'
        }
        sub_resources = {
            'event': Event
        }


class MarketTypeResult(BaseResource):
    """
    :type market_count: int
    :type market_type: unicode
    """
    market_count = None
    market_type = None

    class Meta(BaseResource.Meta):
        identifier = 'market_type_result'
        attributes = {
            'marketCount': 'market_count',
            'marketType': 'market_type'
        }


class CountryResult(BaseResource):
    """
    :type country_code: unicode
    :type market_count: int
    """
    country_code = None
    market_count = None

    class Meta(BaseResource.Meta):
        identifier = 'country_result'
        attributes = {
            'marketCount': 'market_count',
            'countryCode': 'country_code'
        }


class VenueResult(BaseResource):
    """
    :type market_count: int
    :type venue: unicode
    """
    market_count = None
    venue = None

    class Meta(BaseResource.Meta):
        identifier = 'venue_result'
        attributes = {
            'marketCount': 'market_count',
            'venue': 'venue'
        }


class MarketCatalogueDescription(BaseResource):
    """
    :type betting_type: unicode
    :type bsp_market: bool
    :type clarifications: unicode
    :type discount_allowed: bool
    :type each_way_divisor: float
    :type market_base_rate: float
    :type market_time: datetime.datetime
    :type market_type: unicode
    :type persistence_enabled: bool
    :type regulator: unicode
    :type rules: unicode
    :type rules_has_date: bool
    :type suspend_time: datetime.datetime
    :type turn_in_play_enabled: bool
    :type wallet: unicode
    """
    betting_type = None
    bsp_market = None
    clarifications = None
    discount_allowed = None
    each_way_divisor = None
    market_base_rate = None
    market_time = None
    market_type = None
    persistence_enabled = None
    regulator = None
    rules = None
    rules_has_date = None
    suspend_time = None
    turn_in_play_enabled = None
    wallet = None

    class Meta(BaseResource.Meta):
        identifier = 'description'
        attributes = {
            'bettingType': 'betting_type',
            'bspMarket': 'bsp_market',
            'discountAllowed': 'discount_allowed',
            'marketBaseRate': 'market_base_rate',
            'marketTime': 'market_time',
            'marketType': 'market_type',
            'persistenceEnabled': 'persistence_enabled',
            'regulator': 'regulator',
            'rules': 'rules',
            'rulesHasDate': 'rules_has_date',
            'suspendTime': 'suspend_time',
            'turnInPlayEnabled': 'turn_in_play_enabled',
            'wallet': 'wallet',
            'eachWayDivisor': 'each_way_divisor',
            'clarifications': 'clarifications'
        }
        datetime_attributes = (
            'marketTime',
            'suspendTime'
        )


class RunnerCatalogue(BaseResource):
    """
    :type handicap: float
    :type metadata: dict
    :type runner_name: unicode
    :type selection_id: int
    :type sort_priority: int
    """
    handicap = None
    metadata = None
    runner_name = None
    selection_id = None
    sort_priority = None

    class Meta(BaseResource.Meta):
        identifier = 'runners'
        attributes = {
            'selectionId': 'selection_id',
            'runnerName': 'runner_name',
            'sortPriority': 'sort_priority',
            'handicap': 'handicap',
            'metadata': 'metadata'
        }


class MarketCatalogue(BaseResource):
    """
    :type competition: Competition
    :type description: MarketCatalogueDescription
    :type event: Event
    :type event_type: EventType
    :type market_id: unicode
    :type market_name: unicode
    :type market_start_time: datetime.datetime
    :type runners: list[RunnerCatalogue]
    :type total_matched: float
    """
    competition = None
    description = None
    event = None
    event_type = None
    market_id = None
    market_name = None
    market_start_time = None
    runners = None
    total_matched = None

    class Meta(BaseResource.Meta):
        identifier = 'market_catalogue'
        attributes = {
            'marketId': 'market_id',
            'marketName': 'market_name',
            'totalMatched': 'total_matched',
            'marketStartTime': 'market_start_time'
        }
        sub_resources = {
            'competition': Competition,
            'event': Event,
            'eventType': EventType,
            'description': MarketCatalogueDescription,
            'runners': RunnerCatalogue
        }
        datetime_attributes = (
            'marketStartTime'
        )


class PriceSize(BaseResource):
    """
    :type price: float
    :type size: float
    """
    price = None
    size = None

    class Meta(BaseResource.Meta):
        identifier = 'price_size'
        attributes = {
            'price': 'price',
            'size': 'size'
        }


class RunnerBookSP(BaseResource):
    """
    :type actual_sp: float
    :type back_stake_taken: list[PriceSize]
    :type far_price: float
    :type lay_liability_taken: list[PriceSize]
    :type near_price: float
    """
    actual_sp = None
    back_stake_taken = None
    far_price = None
    lay_liability_taken = None
    near_price = None

    class Meta(BaseResource.Meta):
        identifier = 'sp'
        attributes = {
            'nearPrice': 'near_price',
            'farPrice': 'far_price',
            'backStakeTaken': 'back_stake_taken',
            'layLiabilityTaken': 'lay_liability_taken',
            'actualSP': 'actual_sp'
        }
        sub_resources = {
            'backStakeTaken': PriceSize,
            'layLiabilityTaken': PriceSize
        }


class RunnerBookEX(BaseResource):
    """
    :type available_to_back: list[PriceSize]
    :type available_to_lay: list[PriceSize]
    :type traded_volume: list[PriceSize]
    """
    available_to_back = None
    available_to_lay = None
    traded_volume = None

    class Meta(BaseResource.Meta):
        identifier = 'ex'
        attributes = {
            'availableToBack': 'available_to_back',
            'availableToLay': 'available_to_lay',
            'tradedVolume': 'traded_volume'
        }
        sub_resources = {
            'availableToBack': PriceSize,
            'availableToLay': PriceSize,
            'tradedVolume': PriceSize
        }


class RunnerBookOrder(BaseResource):
    """
    :type avg_price_matched: float
    :type bet_id: unicode
    :type bsp_liability: float
    :type order_type: unicode
    :type persistence_type: unicode
    :type placed_date: datetime.datetime
    :type price: float
    :type side: unicode
    :type size: float
    :type size_cancelled: float
    :type size_lapsed: float
    :type size_matched: float
    :type size_remaining: float
    :type size_voided: float
    :type status: unicode
    """
    avg_price_matched = None
    bet_id = None
    bsp_liability = None
    order_type = None
    persistence_type = None
    placed_date = None
    price = None
    side = None
    size = None
    size_cancelled = None
    size_lapsed = None
    size_matched = None
    size_remaining = None
    size_voided = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'orders'
        attributes = {
            'betId': 'bet_id',
            'avgPriceMatched': 'avg_price_matched',
            'bspLiability': 'bsp_liability',
            'orderType': 'order_type',
            'persistenceType': 'persistence_type',
            'placedDate': 'placed_date',
            'price': 'price',
            'side': 'side',
            'size': 'size',
            'sizeCancelled': 'size_cancelled',
            'sizeLapsed': 'size_lapsed',
            'sizeMatched': 'size_matched',
            'sizeRemaining': 'size_remaining',
            'sizeVoided': 'size_voided',
            'status': 'status'
        }
        datetime_attributes = (
            'placedDate'
        )


class RunnerBookMatch(BaseResource):
    """
    :type bet_id: unicode
    :type match_date: datetime.datetime
    :type match_id: unicode
    :type price: float
    :type side: unicode
    :type size: float
    """
    bet_id = None
    match_date = None
    match_id = None
    price = None
    side = None
    size = None

    class Meta(BaseResource.Meta):
        identifier = 'matches'
        attributes = {
            'betId': 'bet_id',
            'matchId': 'match_id',
            'price': 'price',
            'side': 'side',
            'size': 'size',
            'matchDate': 'match_date'
        }
        datetime_attributes = (
            'matchDate'
        )


class RunnerBook(BaseResource):
    """
    :type adjustment_factor: float
    :type ex: RunnerBookEX
    :type handicap: float
    :type last_price_traded: float
    :type matches: list[RunnerBookMatch]
    :type orders: list[RunnerBookOrder]
    :type removal_date: datetime.datetime
    :type selection_id: int
    :type sp: RunnerBookSP
    :type status: unicode
    :type total_matched: float
    """
    adjustment_factor = None
    ex = None
    handicap = None
    last_price_traded = None
    matches = None
    orders = None
    removal_date = None
    selection_id = None
    sp = None
    status = None
    total_matched = None

    class Meta(BaseResource.Meta):
        identifier = 'runners'
        attributes = {
            'selectionId': 'selection_id',
            'status': 'status',
            'totalMatched': 'total_matched',
            'adjustmentFactor': 'adjustment_factor',
            'handicap': 'handicap',
            'lastPriceTraded': 'last_price_traded',
            'removalDate': 'removal_date',
        }
        sub_resources = {
            'sp': RunnerBookSP,
            'ex': RunnerBookEX,
            'orders': RunnerBookOrder,
            'matches': RunnerBookMatch
        }
        datetime_attributes = (
            'removalDate',
        )


class MarketBook(BaseResource):
    """
    :type bet_delay: int
    :type bsp_reconciled: bool
    :type complete: bool
    :type cross_matching: bool
    :type inplay: bool
    :type is_market_data_delayed: bool
    :type last_match_time: datetime.datetime
    :type market_id: unicode
    :type number_of_active_runners: int
    :type number_of_runners: int
    :type number_of_winners: int
    :type publish_time: datetime.datetime
    :type runners: list[RunnerBook]
    :type runners_voidable: bool
    :type status: unicode
    :type total_available: float
    :type total_matched: float
    :type version: int
    """
    bet_delay = None
    bsp_reconciled = None
    complete = None
    cross_matching = None
    inplay = None
    is_market_data_delayed = None
    last_match_time = None
    market_id = None
    number_of_active_runners = None
    number_of_runners = None
    number_of_winners = None
    publish_time = None
    runners = None
    runners_voidable = None
    status = None
    total_available = None
    total_matched = None
    version = None

    class Meta(BaseResource.Meta):
        identifier = 'market_book'
        attributes = {
            'marketId': 'market_id',
            'betDelay': 'bet_delay',
            'bspReconciled': 'bsp_reconciled',
            'complete': 'complete',
            'crossMatching': 'cross_matching',
            'inplay': 'inplay',
            'isMarketDataDelayed': 'is_market_data_delayed',
            'lastMatchTime': 'last_match_time',
            'numberOfActiveRunners': 'number_of_active_runners',
            'numberOfRunners': 'number_of_runners',
            'numberOfWinners': 'number_of_winners',
            'runnersVoidable': 'runners_voidable',
            'status': 'status',
            'totalAvailable': 'total_available',
            'totalMatched': 'total_matched',
            'version': 'version',
            'publishTime': 'publish_time',
        }
        sub_resources = {
            'runners': RunnerBook
        }
        datetime_attributes = (
            'lastMatchTime',
            'publishTime',
        )


class CurrentOrder(BaseResource):
    """
    :type average_price_matched: float
    :type bet_id: unicode
    :type bsp_liability: float
    :type customer_order_ref: unicode
    :type customer_strategy_ref: unicode
    :type handicap: float
    :type market_id: unicode
    :type matched_date: datetime.datetime
    :type order_type: unicode
    :type persistence_type: unicode
    :type placed_date: datetime.datetime
    :type price_size: PriceSize
    :type regulator_code: unicode
    :type selection_id: int
    :type side: unicode
    :type size_cancelled: float
    :type size_lapsed: float
    :type size_matched: float
    :type size_remaining: float
    :type size_voided: float
    :type status: unicode
    """
    average_price_matched = None
    bet_id = None
    bsp_liability = None
    customer_order_ref = None
    customer_strategy_ref = None
    handicap = None
    market_id = None
    matched_date = None
    order_type = None
    persistence_type = None
    placed_date = None
    price_size = None
    regulator_code = None
    selection_id = None
    side = None
    size_cancelled = None
    size_lapsed = None
    size_matched = None
    size_remaining = None
    size_voided = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'orders'
        attributes = {
            'betId': 'bet_id',
            'averagePriceMatched': 'average_price_matched',
            'bspLiability': 'bsp_liability',
            'handicap': 'handicap',
            'marketId': 'market_id',
            'matchedDate': 'matched_date',
            'orderType': 'order_type',
            'persistenceType': 'persistence_type',
            'placedDate': 'placed_date',
            'regulatorCode': 'regulator_code',
            'selectionId': 'selection_id',
            'side': 'side',
            'sizeCancelled': 'size_cancelled',
            'sizeLapsed': 'size_lapsed',
            'sizeMatched': 'size_matched',
            'sizeRemaining': 'size_remaining',
            'sizeVoided': 'size_voided',
            'status': 'status',
            'customerStrategyRef': 'customer_strategy_ref',
            'customerOrderRef': 'customer_order_ref'
        }
        sub_resources = {
            'priceSize': PriceSize
        }
        datetime_attributes = (
            'placedDate',
            'matchedDate'
        )


class CurrentOrders(BaseResource):
    """
    :type more_available: bool
    :type orders: list[CurrentOrder]
    """
    more_available = None
    orders = None

    class Meta(BaseResource.Meta):
        identifier = 'current_orders'
        attributes = {
            'moreAvailable': 'more_available'
        }
        sub_resources = {
            'currentOrders': CurrentOrder
        }


class ClearedOrder(BaseResource):
    """
    :type bet_count: int
    :type bet_id: unicode
    :type bet_outcome: unicode
    :type customer_order_ref: unicode
    :type customer_strategy_ref: unicode
    :type event_id: unicode
    :type event_type_id: unicode
    :type handicap: float
    :type last_matched_date: datetime.datetime
    :type market_id: unicode
    :type order_type: unicode
    :type persistence_type: unicode
    :type placed_date: datetime.datetime
    :type price_matched: float
    :type price_reduced: bool
    :type price_requested: float
    :type profit: float
    :type selection_id: int
    :type settled_date: datetime.datetime
    :type side: unicode
    :type size_settled: float
    """
    bet_count = None
    bet_id = None
    bet_outcome = None
    customer_order_ref = None
    customer_strategy_ref = None
    event_id = None
    event_type_id = None
    handicap = None
    last_matched_date = None
    market_id = None
    order_type = None
    persistence_type = None
    placed_date = None
    price_matched = None
    price_reduced = None
    price_requested = None
    profit = None
    selection_id = None
    settled_date = None
    side = None
    size_settled = None

    class Meta(BaseResource.Meta):
        identifier = 'orders'
        attributes = {
            'betId': 'bet_id',
            'betCount': 'bet_count',
            'betOutcome': 'bet_outcome',
            'eventId': 'event_id',
            'eventTypeId': 'event_type_id',
            'handicap': 'handicap',
            'lastMatchedDate': 'last_matched_date',
            'marketId': 'market_id',
            'orderType': 'order_type',
            'persistenceType': 'persistence_type',
            'placedDate': 'placed_date',
            'priceMatched': 'price_matched',
            'priceReduced': 'price_reduced',
            'priceRequested': 'price_requested',
            'profit': 'profit',
            'selectionId': 'selection_id',
            'settledDate': 'settled_date',
            'side': 'side',
            'sizeSettled': 'size_settled',
            'customerStrategyRef': 'customer_strategy_ref',
            'customerOrderRef': 'customer_order_ref'
        }
        datetime_attributes = (
            'placedDate',
            'lastMatchedDate',
            'settledDate'
        )


class ClearedOrders(BaseResource):
    """
    :type more_available: bool
    :type orders: list[ClearedOrder]
    """
    more_available = None
    orders = None

    class Meta(BaseResource.Meta):
        identifier = 'cleared_orders'
        attributes = {
            'moreAvailable': 'more_available'
        }
        sub_resources = {
            'clearedOrders': ClearedOrder
        }


class ProfitAndLosses(BaseResource):
    """
    :type if_lose: float
    :type if_place: float
    :type if_win: float
    :type selection_id: int
    """
    if_lose = None
    if_place = None
    if_win = None
    selection_id = None

    class Meta(BaseResource.Meta):
        identifier = 'profit_and_losses'
        attributes = {
            'selectionId': 'selection_id',
            'ifWin': 'if_win',
            'ifLose': 'if_lose',
            'ifPlace': 'if_place'
        }


class MarketProfitLoss(BaseResource):
    """
    :type commission_applied: float
    :type market_id: unicode
    :type profit_and_losses: list[ProfitAndLosses]
    """
    commission_applied = None
    market_id = None
    profit_and_losses = None

    class Meta(BaseResource.Meta):
        identifier = 'market_profit_loss'
        attributes = {
            'marketId': 'market_id',
            'commissionApplied': 'commission_applied'
        }
        sub_resources = {
            'profitAndLosses': ProfitAndLosses
        }


class LimitOrder(BaseResource):
    """
    :type bet_target_size: float
    :type bet_target_type: unicode
    :type min_fill_size: float
    :type persistence_type: unicode
    :type price: float
    :type size: float
    :type time_in_force: unicode
    """
    bet_target_size = None
    bet_target_type = None
    min_fill_size = None
    persistence_type = None
    price = None
    size = None
    time_in_force = None

    class Meta(BaseResource.Meta):
        identifier = 'order'
        attributes = {
            'persistenceType': 'persistence_type',
            'price': 'price',
            'size': 'size',
            'timeInForce': 'time_in_force',
            'minFillSize': 'min_fill_size',
            'betTargetType': 'bet_target_type',
            'betTargetSize': 'bet_target_size'
        }


class PlaceOrderInstruction(BaseResource):
    """
    :type customer_order_ref: unicode
    :type handicap: float
    :type order: LimitOrder
    :type order_type: unicode
    :type selection_id: int
    :type side: unicode
    """
    customer_order_ref = None
    handicap = None
    order = None
    order_type = None
    selection_id = None
    side = None

    class Meta(BaseResource.Meta):
        identifier = 'instruction'
        attributes = {
            'selectionId': 'selection_id',
            'side': 'side',
            'orderType': 'order_type',
            'handicap': 'handicap',
            'customerOrderRef': 'customer_order_ref'
        }
        sub_resources = {
            'limitOrder': LimitOrder
        }


class PlaceOrderInstructionReports(BaseResource):
    """
    :type average_price_matched: float
    :type bet_id: unicode
    :type error_code: str
    :type instruction: PlaceOrderInstruction
    :type order_status: unicode
    :type placed_date: datetime.datetime
    :type size_matched: float
    :type status: unicode
    """
    average_price_matched = None
    bet_id = None
    error_code = None
    instruction = None
    order_status = None
    placed_date = None
    size_matched = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'place_instruction_reports'
        attributes = {
            'status': 'status',
            'orderStatus': 'order_status',
            'betId': 'bet_id',
            'averagePriceMatched': 'average_price_matched',
            'sizeMatched': 'size_matched',
            'placedDate': 'placed_date',
            'errorCode': 'error_code',
        }
        sub_resources = {
            'instruction': PlaceOrderInstruction
        }
        datetime_attributes = (
            'placedDate'
        )


class PlaceOrders(BaseResource):
    """
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type place_instruction_reports: list[PlaceOrderInstructionReports]
    :type status: unicode
    """
    customer_ref = None
    error_code = None
    market_id = None
    place_instruction_reports = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'place_orders'
        attributes = {
            'marketId': 'market_id',
            'status': 'status',
            'customerRef': 'customer_ref',
            'errorCode': 'error_code'
        }
        sub_resources = {
            'instructionReports': PlaceOrderInstructionReports
        }


class CancelOrderInstruction(BaseResource):
    """
    :type bet_id: unicode
    :type size_reduction: float
    """
    bet_id = None
    size_reduction = None

    class Meta(BaseResource.Meta):
        identifier = 'instruction'
        attributes = {
            'betId': 'bet_id',
            'sizeReduction': 'size_reduction'
        }


class CancelOrderInstructionReports(BaseResource):
    """
    :type cancelled_date: datetime.datetime
    :type error_code: str
    :type instruction: CancelOrderInstruction
    :type size_cancelled: float
    :type status: unicode
    """
    cancelled_date = None
    error_code = None
    instruction = None
    size_cancelled = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'cancel_instruction_reports'
        attributes = {
            'status': 'status',
            'sizeCancelled': 'size_cancelled',
            'cancelledDate': 'cancelled_date',
            'errorCode': 'error_code',
        }
        sub_resources = {
            'instruction': CancelOrderInstruction
        }
        datetime_attributes = (
            'cancelledDate'
        )


class CancelOrders(BaseResource):
    """
    :type cancel_instruction_reports: list[CancelOrderInstructionReports]
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type status: unicode
    """
    cancel_instruction_reports = None
    customer_ref = None
    error_code = None
    market_id = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'cancel_orders'
        attributes = {
            'marketId': 'market_id',
            'status': 'status',
            'customerRef': 'customer_ref',
            'errorCode': 'error_code'
        }
        sub_resources = {
            'instructionReports': CancelOrderInstructionReports
        }


class UpdateOrderInstruction(BaseResource):
    """
    :type bet_id: unicode
    :type new_persistence_type: unicode
    """
    bet_id = None
    new_persistence_type = None

    class Meta(BaseResource.Meta):
        identifier = 'instruction'
        attributes = {
            'betId': 'bet_id',
            'newPersistenceType': 'new_persistence_type'
        }


class UpdateOrderInstructionReports(BaseResource):
    """
    :type error_code: str
    :type instruction: UpdateOrderInstruction
    :type status: unicode
    """
    error_code = None
    instruction = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'update_instruction_reports'
        attributes = {
            'status': 'status',
            'errorCode': 'error_code',
        }
        sub_resources = {
            'instruction': UpdateOrderInstruction
        }


class UpdateOrders(BaseResource):
    """
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type status: unicode
    :type update_instruction_reports: list[UpdateOrderInstructionReports]
    """
    customer_ref = None
    error_code = None
    market_id = None
    status = None
    update_instruction_reports = None

    class Meta(BaseResource.Meta):
        identifier = 'update_orders'
        attributes = {
            'marketId': 'market_id',
            'status': 'status',
            'customerRef': 'customer_ref',
            'errorCode': 'error_code'
        }
        sub_resources = {
            'instructionReports': UpdateOrderInstructionReports
        }


class ReplaceOrderInstructionReports(BaseResource):
    """
    :type cancel_instruction_reports: CancelOrderInstructionReports
    :type error_code: str
    :type place_instruction_reports: PlaceOrderInstructionReports
    :type status: unicode
    """
    cancel_instruction_reports = None
    error_code = None
    place_instruction_reports = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'replace_instruction_reports'
        attributes = {
            'status': 'status',
            'errorCode': 'error_code',
        }
        sub_resources = {
            'cancelInstructionReport': CancelOrderInstructionReports,
            'placeInstructionReport': PlaceOrderInstructionReports
        }


class ReplaceOrders(BaseResource):
    """
    :type customer_ref: unicode
    :type error_code: str
    :type market_id: unicode
    :type replace_instruction_reports: list[ReplaceOrderInstructionReports]
    :type status: unicode
    """
    customer_ref = None
    error_code = None
    market_id = None
    replace_instruction_reports = None
    status = None

    class Meta(BaseResource.Meta):
        identifier = 'replace_orders'
        attributes = {
            'marketId': 'market_id',
            'status': 'status',
            'customerRef': 'customer_ref',
            'errorCode': 'error_code'
        }
        sub_resources = {
            'instructionReports': ReplaceOrderInstructionReports
        }



from .baseresource import BaseResource


class AccountFunds(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'account_funds'
        attributes = {
            'availableToBetBalance': 'available_to_bet_balance',
            'discountRate': 'discount_rate',
            'exposure': 'exposure',
            'exposureLimit': 'exposure_limit',
            'pointsBalance': 'points_balance',
            'retainedCommission': 'retained_commission',
            'wallet': 'wallet'
        }


class AccountDetails(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'account_details'
        attributes = {
            'countryCode': 'country_code',
            'currencyCode': 'currency_code',
            'discountRate': 'discount_rate',
            'firstName': 'first_name',
            'lastName': 'last_name',
            'localeCode': 'locale_code',
            'pointsBalance': 'points_balance',
            'region': 'region',
            'timezone': 'timezone'
        }


class LegacyData(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'legacy_data'
        attributes = {
            'avgPrice': 'avg_price',
            'betCategoryType': 'bet_category_type',
            'betSize': 'bet_size',
            'betType': 'bet_type',
            'eventId': 'event_id',
            'eventTypeId': 'event_type_id',
            'fullMarketName': 'full_market_name',
            'grossBetAmount': 'gross_bet_amount'
        }


class AccountStatement(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'account_statement'
        attributes = {
            'amount': 'amount',
            'balance': 'balance',
            'itemClass': 'item_class',
            'itemClassData': 'item_class_data',
            'itemDate': 'item_date',
            'refId': 'ref_id'
        }
        sub_resources = {
            'legacyData': LegacyData
        }
        datetime_attributes = (
            'itemDate',
        )


class AccountStatementResult(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'account_statement_result'
        attributes = {
            'moreAvailable': 'more_available'
        }
        sub_resources = {
            'accountStatement': AccountStatement
        }


class CurrencyRate(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'currency_rate'
        attributes = {
            'currencyCode': 'currency_code',
            'rate': 'rate'
        }


class TransferFunds(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'transfer_funds'
        attributes = {
            'transactionId': 'transaction_id'
        }



from .baseresource import BaseResource


class RaceType(BaseResource):
    """
    :type abbr: unicode
    :type full: unicode
    :type key: unicode
    """
    abbr = None
    full = None
    key = None

    class Meta(BaseResource.Meta):
        identifier = 'race_type'
        attributes = {
            'abbr': 'abbr',
            'full': 'full',
            'key': 'key'
        }


class RaceClassification(BaseResource):
    """
    :type classification: unicode
    :type classification_abbr: unicode
    :type code: unicode
    :type display_name: unicode
    :type display_name_abbr: unicode
    """
    classification = None
    classification_abbr = None
    code = None
    display_name = None
    display_name_abbr = None

    class Meta(BaseResource.Meta):
        identifier = 'race_classification'
        attributes = {
            'classification': 'classification',
            'classificationAbbr': 'classification_abbr',
            'code': 'code',
            'displayName': 'display_name',
            'displayNameAbbr': 'display_name_abbr'
        }


class Market(BaseResource):
    """
    :type market_id: unicode
    :type market_type: unicode
    :type number_of_winners: int
    """
    market_id = None
    market_type = None
    number_of_winners = None

    class Meta(BaseResource.Meta):
        identifier = 'markets'
        attributes = {
            'marketId': 'market_id',
            'marketType': 'market_type',
            'numberOfWinners': 'number_of_winners'
        }


class Going(BaseResource):
    """
    :type abbr: unicode
    :type full: unicode
    :type key: unicode
    """
    abbr = None
    full = None
    key = None

    class Meta(BaseResource.Meta):
        identifier = 'going'
        attributes = {
            'abbr': 'abbr',
            'full': 'full',
            'key': 'key'
        }


class Course(BaseResource):
    """
    :type country: unicode
    :type country_code: unicode
    :type course_id: unicode
    :type course_type: unicode
    :type name: unicode
    :type surface_type: unicode
    :type timeform_course_code: unicode
    :type timezone: unicode
    """
    country = None
    country_code = None
    course_id = None
    course_type = None
    name = None
    surface_type = None
    timeform_course_code = None
    timezone = None

    class Meta(BaseResource.Meta):
        identifier = 'course'
        attributes = {
            'country': 'country',
            'countryCode': 'country_code',
            'courseId': 'course_id',
            'courseType': 'course_type',
            'name': 'name',
            'surfaceType': 'surface_type',
            'timeformCourseCode': 'timeform_course_code',
            'timezone': 'timezone'
        }


class Race(BaseResource):
    """
    :type betfair_meeting_id: unicode
    :type course: Course
    :type distance: int
    :type eligibility: unicode
    :type going: Going
    :type is_results_available: bool
    :type markets: list[Market]
    :type meeting_going: unicode
    :type meeting_id: unicode
    :type number_of_runners: int
    :type race_class: int
    :type race_classification: RaceClassification
    :type race_id: unicode
    :type race_title: unicode
    :type race_type: RaceType
    :type start_date: datetime.datetime
    """
    betfair_meeting_id = None
    course = None
    distance = None
    eligibility = None
    going = None
    is_results_available = None
    markets = None
    meeting_going = None
    meeting_id = None
    number_of_runners = None
    race_class = None
    race_classification = None
    race_id = None
    race_title = None
    race_type = None
    start_date = None

    class Meta(BaseResource.Meta):
        identifier = 'race'
        attributes = {
            'betfairMeetingId': 'betfair_meeting_id',
            'distance': 'distance',
            'eligibility': 'eligibility',
            'isResultAvailable': 'is_results_available',
            'meetingGoing': 'meeting_going',
            'meetingId': 'meeting_id',
            'numberOfRunners': 'number_of_runners',
            'raceClass': 'race_class',
            'raceId': 'race_id',
            'raceTitle': 'race_title',
            'startDate': 'start_date',
        }
        sub_resources = {
            'course': Course,
            'going': Going,
            'markets': Market,
            'raceClassification': RaceClassification,
            'raceType': RaceType,
        }
        datetime_attributes = (
            'startDate',
        )


class InPlayHint(BaseResource):
    """
    :type hint_name: unicode
    :type hint_value: unicode
    """
    hint_name = None
    hint_value = None

    class Meta(BaseResource.Meta):
        identifier = 'in_play_hints'
        attributes = {
            'hintName': 'hint_name',
            'hintValue': 'hint_value',
        }


class DaysSinceLastRun(BaseResource):
    """
    :type days: int
    :type type: unicode
    """
    days = None
    type = None

    class Meta(BaseResource.Meta):
        identifier = 'days_since_last_run'
        attributes = {
            'days': 'days',
            'type': 'type'
        }


class Jockey(BaseResource):
    """
    :type jockey_id: unicode
    :type name: unicode
    """
    jockey_id = None
    name = None

    class Meta(BaseResource.Meta):
        identifier = 'jockey'
        attributes = {
            'jockeyId': 'jockey_id',
            'name': 'name'
        }


class Selection(BaseResource):
    """
    :type market_id: unicode
    :type market_type: unicode
    :type selection_id: unicode
    """
    market_id = None
    market_type = None
    selection_id = None

    class Meta(BaseResource.Meta):
        identifier = 'selections'
        attributes = {
            'marketId': 'market_id',
            'marketType': 'market_type',
            'selectionId': 'selection_id'
        }


class Trainer(BaseResource):
    """
    :type location: unicode
    :type name: unicode
    :type trainer_id: unicode
    """
    location = None
    name = None
    trainer_id = None

    class Meta(BaseResource.Meta):
        identifier = 'trainer'
        attributes = {
            'location': 'location',
            'name': 'name',
            'trainerId': 'trainer_id'
        }


class Wearing(BaseResource):
    """
    :type abbr: unicode
    :type full: unicode
    :type key: unicode
    """
    abbr = None
    full = None
    key = None

    class Meta(BaseResource.Meta):
        identifier = 'wearing'
        attributes = {
            'abbr': 'abbr',
            'full': 'full',
            'key': 'key'
        }


class Runner(BaseResource):
    """
    :type age: int
    :type comment: unicode
    :type days_since_last_run: DaysSinceLastRun
    :type draw: int
    :type gender: unicode
    :type horse_id: unicode
    :type is_non_runner: bool
    :type jockey: Jockey
    :type long_handicap: int
    :type name: unicode
    :type official_rating: int
    :type owner_colours: unicode
    :type recent_form: unicode
    :type saddle_cloth: unicode
    :type selections: list[Selection]
    :type star_rating: int
    :type timeform_123_place: int
    :type trainer: Trainer
    :type wearing: Wearing
    :type weight: int
    :type wins_at_course: int
    :type wins_at_course_and_distance: int
    :type wins_at_distance: int
    """
    age = None
    comment = None
    days_since_last_run = None
    draw = None
    gender = None
    horse_id = None
    is_non_runner = None
    jockey = None
    long_handicap = None
    name = None
    official_rating = None
    owner_colours = None
    recent_form = None
    saddle_cloth = None
    selections = None
    star_rating = None
    timeform_123_place = None
    trainer = None
    wearing = None
    weight = None
    wins_at_course = None
    wins_at_course_and_distance = None
    wins_at_distance = None

    class Meta(BaseResource.Meta):
        identifier = 'runners'
        attributes = {
            'age': 'age',
            'comment': 'comment',
            'gender': 'gender',
            'horseId': 'horse_id',
            'isNonRunner': 'is_non_runner',
            'longHandicap': 'long_handicap',
            'name': 'name',
            'ownerColours': 'owner_colours',
            'recentForm': 'recent_form',
            'saddleCloth': 'saddle_cloth',
            'starRating': 'star_rating',
            'timeform123Place': 'timeform_123_place',
            'weight': 'weight',
            'winsAtCourse': 'wins_at_course',
            'winsAtCourseAndDistance': 'wins_at_course_and_distance',
            'winsAtDistance': 'wins_at_distance',
            'draw': 'draw',
            'officialRating': 'official_rating'
        }
        sub_resources = {
            'daysSinceLastRun': DaysSinceLastRun,
            'jockey': Jockey,
            'selections': Selection,
            'trainer': Trainer,
            'wearing': Wearing
        }


class RaceCard(BaseResource):
    """
    :type betting_forecast_text: unicode
    :type comment: unicode
    :type in_play_hints: InPlayHint
    :type minimum_weight: int
    :type number_of_non_runners: int
    :type number_of_runners: int
    :type prize: unicode
    :type race: Race
    :type runners: list[Runner]
    :type timeform_123_text: unicode
    """
    betting_forecast_text = None
    comment = None
    in_play_hints = None
    minimum_weight = None
    number_of_non_runners = None
    number_of_runners = None
    prize = None
    race = None
    runners = None
    timeform_123_text = None

    class Meta(BaseResource.Meta):
        identifier = 'race_card'
        attributes = {
            'bettingForecastText': 'betting_forecast_text',
            'comment': 'comment',
            'minimumWeight': 'minimum_weight',
            'numberOfNonRunners': 'number_of_non_runners',
            'numberOfRunners': 'number_of_runners',
            'prize': 'prize',
            'timeform123Text': 'timeform_123_text'
        }
        sub_resources = {
            'inPlayHints': InPlayHint,
            'race': Race,
            'runners': Runner
        }



import datetime

from .baseendpoint import BaseEndpoint
from .. import resources


class Scores(BaseEndpoint):

    URI = 'ScoresAPING/v1.0/'

    def list_race_details(self, params=None, session=None):
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listRaceDetails')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.RaceDetails, date_time_sent)

    @property
    def url(self):
        return '%s%s' % (self.client.api_uri, 'scores/json-rpc/v1')



import datetime

from .enums import MockParams
from .exceptions import StatusCodeError


def check_status_code(response, codes=None):
    """Checks response.status_code is in codes

    :param response: Requests response
    :param codes: List of accepted codes or callable
    :raises: StatusCodeError if code invalid
    """
    codes = codes or [200]
    if response.status_code not in codes:
        raise StatusCodeError(response.status_code)


# def api_request(func):
#     """Checks params and provides MockParams if None
#
#     :param func: api request function.
#     """
#     api_request_name = func.__name__
#
#     def _api_request(api, params=None, session=None, exchange=None):
#         if not api.check_session:
#             raise SessionTokenError()
#         if not params:
#             params = MockParams[api_request_name].value
#             return func(api, params, session, exchange)
#         else:
#             return func(api, params, session, exchange)
#     return _api_request


def strp_betfair_time(datetime_string):
    """Converts Betfair string to datetime.

    :param datetime_string: Datetime string.
    """
    try:
        return datetime.datetime.strptime(datetime_string, "%Y-%m-%dT%H:%M:%S.%fZ")
    except TypeError:
        return None
    except ValueError:
        return None


def strp_betfair_integer_time(datetime_integer):
    """Converts Betfair integer to utc datetime.

    :param datetime_integer: Datetime integer.
    """
    try:
        return datetime.datetime.utcfromtimestamp(datetime_integer / 1e3)
    except TypeError:
        return None


def price_check(data, number):
    """Access price data from dictionary.

    :param data: Dict object.
    :param number: Number.
    """
    try:
        output = data[number].price
    except KeyError:
        output = None
    except IndexError:
        output = None
    return output


def size_check(data, number):
    """Access size data from dictionary.

    :param data: Dict object.
    :param number: Number.
    """
    try:
        output = data[number].size
    except KeyError:
        output = None
    except IndexError:
        output = None
    return output


def update_available(available, book_update, deletion_select):
    for book in book_update:
        updated = False
        for (count, trade) in enumerate(available):
            if trade[0] == book[0]:
                if book[deletion_select] == 0:
                    del available[count]
                    updated = True
                    break
                else:
                    available[count] = book
                    updated = True
                    break

        if not updated and book[deletion_select] != 0:
            # handles betfair bug, http://forum.bdp.betfair.com/showthread.php?t=3351
            available.append(book)



from ..streaming import BaseListener, BetfairStream


class Streaming(object):

    def __init__(self, parent):
        """
        :param parent: API client.
        """
        self.client = parent

    def create_stream(self, unique_id, listener=None, timeout=11, buffer_size=1024, description='BetfairSocket'):
        listener = listener if listener else BaseListener()
        return BetfairStream(unique_id, listener, app_key=self.client.app_key, session_token=self.client.session_token,
                             timeout=timeout, buffer_size=buffer_size, description=description)



import json
from requests import ConnectionError

from ..exceptions import APIError
from ..utils import check_status_code


class BaseEndpoint(object):

    connect_timeout = 3.05
    read_timeout = 16
    _error = APIError

    def __init__(self, parent):
        """
        :param parent: API client.
        """
        self.client = parent

    def request(self, method, params, session):
        """
        :param method: Betfair api-ng method to be used.
        :param params: Params to be used in request, if None will use MockParams Enum.  # todo
        :param session: Requests session to be used, reduces latency.
        """
        session = session or self.client.session
        request = self.create_req(method, params)
        try:
            response = session.post(self.url, data=request, headers=self.client.request_headers,
                                    timeout=(self.connect_timeout, self.read_timeout))
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)

        check_status_code(response)
        if self._error_handler:
            self._error_handler(response.json(), method, params)
        return response

    @staticmethod
    def create_req(method, params):
        """
        :param method: Betfair api-ng method to be used.
        :param params: Params to be used in request.
        :return: Json payload.
        """
        payload = {'jsonrpc': '2.0',
                   'method': method,
                   'params': params,
                   'id': 1}
        return json.dumps(payload)

    def _error_handler(self, response, method=None, params=None):
        """
        :param response: Json response.
        :param params: Params to be used in request.
        :param method: Betfair api-ng method to be used.
        :return: None if no error or _error raised.
        """
        if response.get('result'):
            return
        elif response.get('error'):
            raise self._error(response, method, params)

    @staticmethod
    def process_response(response_json, resource, date_time_sent):
        """
        :param response_json: Response in json format
        :param resource: Resource data structure
        :param date_time_sent: Date time sent
        """
        if isinstance(response_json, list):
            return [resource(date_time_sent=date_time_sent, **x) for x in response_json]
        else:
            response_result = response_json.get('result', response_json)
            if isinstance(response_result, list):
                return [resource(date_time_sent=date_time_sent, **x) for x in response_result]
            else:
                return resource(date_time_sent=date_time_sent, **response_result)

    @property
    def url(self):
        return '%s%s' % (self.client.api_uri, 'betting/json-rpc/v1')



from .baseresource import BaseResource


class RaceDetails(BaseResource):
    class Meta(BaseResource.Meta):
        identifier = 'race_details'
        attributes = {
            'responseCode': 'response_code',
            'raceId': 'race_id',
            'meetingId': 'meeting_id',
            'raceStatus': 'race_status',
            'lastUpdated': 'last_updated'
        }
        datetime_attributes = (
            'lastUpdated'
        )



import datetime

from .baseendpoint import BaseEndpoint
from .. import resources


class Betting(BaseEndpoint):

    URI = 'SportsAPING/v1.0/'

    def list_event_types(self, params=None, session=None):
        """
        :rtype: list[resources.EventTypeResult]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listEventTypes')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.EventTypeResult, date_time_sent)

    def list_competitions(self, params=None, session=None):
        """
        :rtype: list[resources.CompetitionResult]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listCompetitions')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.CompetitionResult, date_time_sent)

    def list_time_ranges(self, params=None, session=None):
        """
        :rtype: list[resources.TimeRangeResult]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listTimeRanges')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.TimeRangeResult, date_time_sent)

    def list_events(self, params=None, session=None):
        """
        :rtype: list[resources.EventResult]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listEvents')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.EventResult, date_time_sent)

    def list_market_types(self, params=None, session=None):
        """
        :rtype: list[resources.MarketTypeResult]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listMarketTypes')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.MarketTypeResult, date_time_sent)

    def list_countries(self, params=None, session=None):
        """
        :rtype: list[resources.CountryResult]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listCountries')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.CountryResult, date_time_sent)

    def list_venues(self, params=None, session=None):
        """
        :rtype: list[resources.VenueResult]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listVenues')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.VenueResult, date_time_sent)

    def list_market_catalogue(self, params=None, session=None):
        """
        :rtype: list[resources.MarketCatalogue]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listMarketCatalogue')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.MarketCatalogue, date_time_sent)

    def list_market_book(self, params=None, session=None):
        """
        :rtype: list[resources.MarketBook]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listMarketBook')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.MarketBook, date_time_sent)

    def list_current_orders(self, params=None, session=None):
        """
        :rtype: list[resources.CurrentOrders]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listCurrentOrders')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.CurrentOrders, date_time_sent)

    def list_cleared_orders(self, params=None, session=None):
        """
        :rtype: list[resources.ClearedOrders]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listClearedOrders')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.ClearedOrders, date_time_sent)

    def list_market_profit_and_loss(self, params=None, session=None):
        """
        :rtype: list[resources.MarketProfitLoss]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'listMarketProfitAndLoss')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.MarketProfitLoss, date_time_sent)

    def place_orders(self, params=None, session=None):
        """
        :rtype: list[resources.PlaceOrders]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'placeOrders')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.PlaceOrders, date_time_sent)

    def cancel_orders(self, params=None, session=None):
        """
        :rtype: list[resources.CancelOrders]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'cancelOrders')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.CancelOrders, date_time_sent)

    def update_orders(self, params=None, session=None):
        """
        :rtype: list[resources.UpdateOrders]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'updateOrders')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.UpdateOrders, date_time_sent)

    def replace_orders(self, params=None, session=None):
        """
        :rtype: list[resources.ReplaceOrders]
        """
        date_time_sent = datetime.datetime.utcnow()
        method = '%s%s' % (self.URI, 'replaceOrders')
        response = self.request(method, params, session)
        return self.process_response(response.json(), resources.ReplaceOrders, date_time_sent)



import datetime
from requests import ConnectionError

from ..exceptions import APIError
from ..utils import check_status_code
from .baseendpoint import BaseEndpoint
from .. import resources


class InPlayService(BaseEndpoint):

    def get_event_timeline(self, event_id, session=None):
        date_time_sent = datetime.datetime.utcnow()
        url = '%s%s' % (self.url, 'eventTimeline')
        params = {
            'eventId': event_id,
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        response = self.request(params=params, session=session, url=url)
        return self.process_response(response.json(), resources.EventTimeline, date_time_sent)

    def get_scores(self, event_ids, session=None):
        date_time_sent = datetime.datetime.utcnow()
        url = '%s%s' % (self.url, 'scores')
        params = {
            'eventIds': ','.join(str(x) for x in event_ids),
            'alt': 'json',
            'regionCode': 'UK',
            'locale': 'en_GB'
        }
        response = self.request(params=params, session=session, url=url)
        return self.process_response(response.json(), resources.Scores, date_time_sent)

    def request(self, method=None, params=None, session=None, url=None):
        session = session or self.client.session
        try:
            response = session.get(url, params=params, headers=self.headers)
        except ConnectionError:
            raise APIError(None, method, params, 'ConnectionError')
        except Exception as e:
            raise APIError(None, method, params, e)

        check_status_code(response)
        return response

    @property
    def headers(self):
        return {
            'Connection': 'keep-alive',
            'Content-Type': 'application/json'
        }

    @property
    def url(self):
        return 'https://www.betfair.com/inplayservice/v1.1/'



from .baseclient import BaseClient
from . import endpoints


class APIClient(BaseClient):

    def __init__(self, username, password=None, app_key=None, certs=None, locale=None, cert_files=None):
        super(APIClient, self).__init__(username, password, app_key=app_key, certs=certs, locale=locale, cert_files=cert_files)

        self.login = endpoints.Login(self)
        self.keep_alive = endpoints.KeepAlive(self)
        self.logout = endpoints.Logout(self)
        self.betting = endpoints.Betting(self)
        self.account = endpoints.Account(self)
        self.navigation = endpoints.Navigation(self)
        self.scores = endpoints.Scores(self)
        self.streaming = endpoints.Streaming(self)
        self.in_play_service = endpoints.InPlayService(self)
        self.race_card = endpoints.RaceCard(self)

    def __repr__(self):
        return '<APIClient [%s]>' % self.username

    def __str__(self):
        return 'APIClient'
